/* 2 September 2001, Alexandros Droseltis */

/* This file includes the necessary functions for the TRANSPOSER,
   and the transposer itself */

void first_step(int interval, char class0, int *class);
void transpositive(int input[][2], int i, int interval, int class, int output[][2]);
int Tr_Condition(int interval, int note);
void transposer(int mel_num[][2], int i, int t, char c2, int trans1[][2]);
void oct_d_trans(int input[][2], int i, int oct, int output[][2]);
void negative(int interval1, char class1, int *interval2, char *class2, int *octave);
void transupper(int input[][2], int i, int interval, char class0, int output[][2]);


/* 
   NAME: first_step

   FUNCTION

   first_step reads the quantity and the quality of the transposition
   interval and sends a number to the function "transpositive". This number (*class)
   enables the function transpositive to make a right transposition of the 
   accidental according to the quality of the transposition interval, by keeping
   for every interval the same algorithm.

   INPUT
   int interval: the transposition interval (quantity, e.g. 4=fourth, 5=fifth etc.)
   char class0 : the "sort" of the transposition interval. It depends on the quantity
                  (fourth, seventh) and the quality (augmented, minor) of it.
 
   OUTPUT
   int *class  : the number that is beiing sent to "transpositive". It defines the
   transposition of the accidental for each case.
*/
void first_step(int interval, char class0, int *class)
{
  int t0;
  char c0;

  t0=interval;
  c0=class0;

  /* Fourth and Octave */
  if((t0+7)%7==4 || (t0+7)%7==1)
    {
      if(c0=='d') *class=0;
      else if(c0=='p') *class=1;
      else if(c0=='a') *class=2;
    }
  /* Fifth */
  else if((t0+7)%7==5)
    {
      if(c0=='d') *class=1;
      else if(c0=='p') *class=2;
      else if(c0=='a') *class=3;
    }
  /* 2nds, 3rds, 6ths, 7ths */
  else
  {
    if(c0=='d') *class=0;
    else if(c0=='a') *class=3;
    else if(c0=='m') *class=1;
    else if(c0=='M') *class=2;
    else
      {
	*class=-1;
	printf("Please check your entry and try again\n\n");
      }
  }
}

/* NAME: transpositive

   FUNCTION
   "transpositive" transposes a note upwards according:
   a) to the note (accidental & name); the relation is defined from the function
   "Tr_Condition".
   b) to a number that depends on the transposition interval (defined from
   the function "first_step").

   first_step                               Tr_condition
   |                                        |
   V                                        V
   class (BEFORE "transpositive")           0 or 1 (IN "transpositive")
   

   if Tr_condition=0 => minor accidental transposition
   if Tr_condition=1 => major accidental transposition

   What is "major" and "minor" accidental transposition?
   
   If a note is being transposed by a certain interval, there are two cases:
   a) the transposed note has an accidental of value x;
   b) the transposed note has an accidental of value x+1;
   That depends on the note that is going to be transposed and the quantity of the 
   transposition interval (fourth, sixth etc.)

   Moreover the accidental of the transposed note depends on the quantity (fourth, seventh)
   and the quality (minor, major, perfect etc.) of the transposition interval.
   This is being defined by the function "first_step".


   INPUT
   int i        : the place of the first dimension at the arrays input[][] and output[][]
   int input[][]: the note. input[i][0]=accidental, input[i][1]=name of the note
   int interval : the transposition interval (quantity, e.g. 4=fourth, 5=fifth etc.)
   int class    : the "sort" of the transposition interval. It depends on the quantity
                  (fourth, seventh) and the quality (augmented, minor) of it, and it is
                  defined from the function "first_step".

  OUTPUT
  int output  : the transposed note. output[i][0]=accidental, output[i][1]=name of the note
*/



void transpositive(int input[][2], int i, int interval, int class, int output[][2])
{
  int note;

      output[i][1]=input[i][1]+(interval-1);
      note=input[i][1];
      
      if(Tr_Condition(interval, note)==0)
	output[i][0]=input[i][0]-2+class;
      else output[i][0]=input[i][0]-1+class;
}

/*
  NAME: Tr_Condition
  FUNCTION
  s. "transpositive"

  INPUT
  int interval
  int note

  OUTPUT
  O or 1 (s. "transpositive")

  explanation of the following contition:

  Return 1, if the accidental transposition is minor:
  -- interval is a second, and the note is neither an "e" nor a "b"
  (the rest notes require raise of the accidental => major accidental transposition);
  -- interval is a third and the note is "c" or "f" or "g";
  -- interval is a fourth and the note is an "f";
  -- interval is a fifth and the note is not a "b";
  -- interval is a sixth and the note is neither an "e" nor an "a" nor a "b";
  -- interval is a seventh and the note is a "c" or an "f";

  Return 0, if the accidental transposition is major (the rest of the cases)
*/

int Tr_Condition(int interval, int note)
{

  if(((interval+7)%7==2 && ((note+7)%7!=2 && (note+7)%7!=6)) || 
     ((interval+7)%7==3 && ((note+7)%7==0 || (note+7)%7==3 || (note+7)%7==4)) ||
     ((interval+7)%7==6 && ((note+7)%7!=2 && (note+7)%7!=5 && (note+7)%7!=6)) ||
     ((interval+7)%7==0 && ((note+7)%7==0 || (note+7)%7==3)) ||
     ((interval+7)%7==4 && (note+7)%7==3) ||
     ((interval+7)%7==5 && (note+7)%7!=6))
    return (0);
  else return(1);
}

/*
  NAME: first_step

  FUNCTION
  "transupper" transposes a note upwards according to the transposition interval.

                                   transposition interval
			    	   int interval, char class
  note                             =======================>         transposed note
  (accidental  note name)                                           (accidental   note name)
  input[i][0], input[i][1]                                          output[i][0], output[i][1]
  
  INPUT
  int i        : the place of the first dimension at the arrays input[][] and output[][]
  int input[][]: the note. input[i][0]=accidental, input[i][1]=name of the note
  int interval : the transposition interval (quantity, e.g. 4=fourth, 5=fifth etc.)
  char class0  : the sort of the transposition interval (p=perfect, M=major, m=minor)

  OUTPUT
  int output  : the transposed note. output[i][0]=accidental, output[i][1]=name of the note
  
  NOTES
  The function "transupper" unites the functions "first_step" and "transpositive".
*/

void transupper(int input[][2], int i, int interval, char class0, int output[][2])
{
  int c;
 
  first_step(interval, class0, &c);
  transpositive(input, i, interval, c, output);
}

/*
  NAME: negative

  FUNCTION
  "negative" returns the complementary (not regards to the number of the
  octaves tha it contains) of an interval. For example, if the
  source interval is -7M, the destination interval will be +2m, if the
  source interval is -10m, the destination interval will be +3M.
  "negative" remembers how many octaves upwards is the interval transformed, so that
  the function "oct_d_trans" will correct later the transposed note to its right
  octave place.

  INPUT
  int interval1 : the transposition interval (quantity)
  char class1   : the quality of the transposition interval

  OUTPUT
  int *interval : the complemetary interval (not regards to the octaves it contains)
  int *class2   : the class of the complemetary interval (...)
  int *octave   : the number of the octaves it was upwards transformed
*/


void negative(int interval1, char class1, int *interval2, char *class2, int *octave)
{
  int i, p;

  p=-1;
  i=0;
  while(p<0)
    {
      p=9+interval1+7*i;
      i++;
      //printf("p=%d, i=%d\n", p, i);
    }

  if(class1=='d') *class2='a';
  else if(class1=='m') *class2='M';
  else if(class1=='M') *class2='m';
  else if(class1=='p') *class2='p';
  else if(class1=='a') *class2='d';

  *interval2=p;
  *octave=i;

  /* Print results for check */
  /*
  printf("\n");
  printf("Old interval=%d%c, New interval=%d%c\n", interval1, class1, *interval2, *class2);
  printf("Octaves=+%d\n", *octave);
  printf("\n");
  */
}

/*
  NAME: oct_d_trans

  FUNCTION
  "oct_d_trans" transposes a note so many octaves downwards as being told.

  INPUT
  int input[][]: the source note (input[i][0], input[i][1], accidental, name of the note)
  int i        : the place at the first dimension in the arrays input[][] and output[][]
  int oct      : how many octaves downwards must be the note transposed

  OUTPUT
  int output[][]: the tranposed note
*/

void oct_d_trans(int input[][2], int i, int oct, int output[][2])
{
  output[i][0]=input[i][0];
  output[i][1]=input[i][1]-7*oct;
}

/*
  NAME: transposer
  FUNCTION
  "transposer" transposes a note according to a given interval.
  If the intervals is upward, the "transupper" makes the transposition.
  If the interval is downward, "negative" gives an alternative interval for the
  transposition, the complementary one (not regards to the number of octaves that
  are contained in the transposition interval)."negative" also finds out the number
  of octaves are needed in order to transform the transposition. It sends the alternative
  interval (it is upward) to the "transupper" that transposes the note upwards according
  to it. Then, "oct_d_trans" transposes the transposed note as many octaves downwards
  as "negative" needed to transform the given transposition interval.

  For example, if the transposition interval is -9m, this procces is followed:

  a) "negative" returns the alternative transposition interval 7M,
     the number of the octaves that have been needed for the transformation, i.d. 2
     (-9m => -2m => 7m).
  b) "transupper" transposes the note a 7M upwards.
  c) "oct_d_trans" transposes the transposed note 2 octaves downwards.
*/
void transposer(int mel_num[][2], int i, int t, char c2, int trans1[][2])
{
  char class2;
  int c, k, trans_temp[12][2], interval2, octave;

  if(t>0) transupper(mel_num, i, t, c2, trans1);
  else
    {
      negative(t, c2, &interval2, &class2, &octave);
      transupper(mel_num, i, interval2, class2, trans_temp);
      oct_d_trans(trans_temp, i, octave, trans1);
    }
}
